a {
    color: #FC6600;
}

a:hover {
    color: #E65D00;
}

a:visited {
    color: #5F2700;
}

.caption-text {
    color: #CF5400;
}

.wy-nav-top {
    background-color: #CF5400;
}

.wy-side-nav-search > a:hover {
    color: white;
}

.wy-side-nav-search > a:visited {
    color: white;
}

.wy-side-nav-search > div.version {
    color: white;
}

.wy-menu-vertical a:visited {
    color: white;
}

.wy-menu-vertical a:active {
    background-color: #CF5400;
}

.rst-content .note .admonition-title {
    background: #FD9D5C;
}

.note.admonition {
    background: #FED5B9;
}